@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 1997-2016. All Rights Reserved.
+%% Copyright Ericsson AB 1997-2021. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -37,6 +37,10 @@
 -export([nslookup/3, nslookup/4]).
 -export([nnslookup/4, nnslookup/5]).
 
+-export_type([res_option/0,
+              res_error/0,
+              nameserver/0]).
+
 -include_lib("kernel/include/inet.hrl").
 -include("inet_res.hrl").
 -include("inet_dns.hrl").
@@ -259,7 +263,7 @@ do_nslookup(Name, Class, Type, Opts, Timeout) ->
 %%
 -record(options, { % These must be sorted!
 	  alt_nameservers,edns,inet6,nameservers,recurse,
-	  retry,timeout,udp_payload_size,usevc,
+	  retry,servfail_retry_timeout,timeout,udp_payload_size,usevc,
 	  verbose}). % this is a local option, not in inet_db
 %%
 %% Opts when is_list(Opts) -> #options{}
@@ -349,9 +353,6 @@ gethostbyaddr_tm({A,B,C,D} = IP, Timer) when ?ip(A,B,C,D) ->
 	{ok, HEnt} -> {ok, HEnt};
 	_ -> res_gethostbyaddr(dn_in_addr_arpa(A,B,C,D), IP, Timer)
     end;
-%% ipv4  only ipv6 address
-gethostbyaddr_tm({0,0,0,0,0,16#ffff,G,H},Timer) when is_integer(G+H) ->
-    gethostbyaddr_tm({G div 256, G rem 256, H div 256, H rem 256},Timer);
 gethostbyaddr_tm({A,B,C,D,E,F,G,H} = IP, Timer) when ?ip6(A,B,C,D,E,F,G,H) ->
     inet_db:res_update_conf(),
     case inet_db:gethostbyaddr(IP) of
@@ -431,28 +432,7 @@ gethostbyname(Name,Family,Timeout) ->
 gethostbyname_tm(Name,inet,Timer) ->
     getbyname_tm(Name,?S_A,Timer);
 gethostbyname_tm(Name,inet6,Timer) ->
-    case getbyname_tm(Name,?S_AAAA,Timer) of
-	{ok,HEnt} -> {ok,HEnt};
-	{error,nxdomain} ->
-	    case getbyname_tm(Name, ?S_A,Timer) of
-		{ok, HEnt} ->
-		    %% rewrite to a ipv4 only ipv6 address
-		    {ok,
-		     HEnt#hostent {
-		       h_addrtype = inet6,
-		       h_length = 16,
-		       h_addr_list = 
-		       lists:map(
-			 fun({A,B,C,D}) ->
-				 {0,0,0,0,0,16#ffff,A*256+B,C*256+D}
-			 end, HEnt#hostent.h_addr_list)
-		      }};
-		Error ->
-		    Error
-	    end;
-	Error ->
-	    Error
-    end;
+    getbyname_tm(Name,?S_AAAA,Timer);
 gethostbyname_tm(_Name, _Family, _Timer) ->
     {error, einval}.
 	    
@@ -715,21 +695,22 @@ udp_send(#sock{inet=I}, {A,B,C,D}=IP, Port, Buffer)
     gen_udp:send(I, IP, Port, Buffer).
 
 udp_recv(#sock{inet6=I}, {A,B,C,D,E,F,G,H}=IP, Port, Timeout, Decode)
-  when ?ip6(A,B,C,D,E,F,G,H), ?port(Port) ->
-    do_udp_recv(I, IP, Port, Timeout, Decode, time_now(), Timeout);
+  when ?ip6(A,B,C,D,E,F,G,H), ?port(Port), 0 =< Timeout ->
+    do_udp_recv(I, IP, Port, Timeout, Decode, time(Timeout), Timeout);
 udp_recv(#sock{inet=I}, {A,B,C,D}=IP, Port, Timeout, Decode)
-  when ?ip(A,B,C,D), ?port(Port) ->
-    do_udp_recv(I, IP, Port, Timeout, Decode, time_now(), Timeout).
+  when ?ip(A,B,C,D), ?port(Port), 0 =< Timeout ->
+    do_udp_recv(I, IP, Port, Timeout, Decode, time(Timeout), Timeout).
 
-do_udp_recv(_I, _IP, _Port, 0, _Decode, _Start, _T) ->
+do_udp_recv(_I, _IP, _Port, 0, _Decode, _Time, PollCnt)
+  when PollCnt =< 0 ->
     timeout;
-do_udp_recv(I, IP, Port, Timeout, Decode, Start, T) ->
-    case gen_udp:recv(I, 0, T) of
+do_udp_recv(I, IP, Port, Timeout, Decode, Time, PollCnt) ->
+    case gen_udp:recv(I, 0, Timeout) of
 	{ok,Reply} ->
 	    case Decode(Reply) of
-		false when T =:= 0 ->
+		false when Timeout =:= 0 ->
 		    %% This is a compromize between the hard way i.e
-		    %% in the clause below if NewT becomes 0 bailout
+		    %% in the clause below if Timeout becomes 0 bailout
 		    %% immediately and risk that the right reply lies
 		    %% ahead after some bad id replies, and the
 		    %% forgiving way i.e go on with Timeout 0 until
@@ -737,15 +718,12 @@ do_udp_recv(I, IP, Port, Timeout, Decode, Start, T) ->
 		    %% which opens for a DOS attack by a malicious
 		    %% DNS server flooding with bad id replies causing
 		    %% an infinite loop here.
-		    %%
-		    %% Timeout is used as a sanity limit counter
-		    %% just to put an end to the loop.
-		    NewTimeout = erlang:max(0, Timeout - 50),
-		    do_udp_recv(I, IP, Port, NewTimeout, Decode, Start, T);
+                    %%
+		    do_udp_recv(
+                      I, IP, Port, Timeout, Decode, Time, PollCnt-50);
 		false ->
-		    Now = time_now(),
-		    NewT = erlang:max(0, Timeout - now_ms(Now, Start)),
-		    do_udp_recv(I, IP, Port, Timeout, Decode, Start, NewT);
+		    do_udp_recv(
+                      I, IP, Port, timeout(Time), Decode, Time, PollCnt);
 		Result ->
 		    Result
 	    end;
@@ -782,92 +760,202 @@ udp_close(#sock{inet=I,inet6=I6}) ->
 %% And that is what the code seems to do, now fixed, hopefully...
 
 do_query(_Q, [], _Timer) ->
+    %% We have no name server to ask, so say nxdomain
     {error,nxdomain};
 do_query(#q{options=#options{retry=Retry}}=Q, NSs, Timer) ->
-    query_retries(Q, NSs, Timer, Retry, 0, #sock{}).
+    %% We have at least one name server,
+    %% so a failure will be a timeout,
+    %% unless a name server says otherwise
+    Reason = timeout,
+    %% Verify that the nameservers list contains only 2-tuples
+    %% to protect our internal servfail_retry mechanism from surprises
+    lists:all(
+      fun (NS) when tuple_size(NS) =:= 2 -> true;
+          (_) -> false
+      end, NSs) orelse
+        erlang:error(badarg, [Q,NSs,Timer]),
+    query_retries(Q, NSs, Timer, Retry, 0, #sock{}, Reason).
 
-query_retries(_Q, _NSs, _Timer, Retry, Retry, S) ->
-    udp_close(S),
-    {error,timeout};
-query_retries(_Q, [], _Timer, _Retry, _I, S) ->
-    udp_close(S),
-    {error,timeout};
-query_retries(Q, NSs, Timer, Retry, I, S0) ->
-    case query_nss(Q, NSs, Timer, Retry, I, S0, []) of
-	{S,{noanswer,ErrNSs}} -> %% remove unreachable nameservers
-	    query_retries(Q, NSs--ErrNSs, Timer, Retry, I+1, S);
-	{S,Result} ->
-	    udp_close(S),
-	    Result
-    end.
+%% Loop until out of retries or name servers
+%%
+query_retries(_Q, _NSs, _Timer, Retry, I, S, Reason) when Retry =:= I ->
+    query_retries_error(S, Reason);
+query_retries(_Q, [], _Timer, _Retry, _I, S, Reason) ->
+    query_retries_error(S, Reason);
+query_retries(Q, NSs, Timer, Retry, I, S, Reason) ->
+    query_nss(Q, NSs, Timer, Retry, I, S, Reason, []).
 
-query_nss(_Q, [], _Timer, _Retry, _I, S, ErrNSs) ->
-    {S,{noanswer,ErrNSs}};
-query_nss(#q{edns=undefined}=Q, NSs, Timer, Retry, I, S, ErrNSs) ->
-    query_nss_dns(Q, NSs, Timer, Retry, I, S, ErrNSs);
-query_nss(Q, NSs, Timer, Retry, I, S, ErrNSs) ->
-    query_nss_edns(Q, NSs, Timer, Retry, I, S, ErrNSs).
+%% For each name server:
+%%     If EDNS is enabled, try that first,
+%%     and for selected failures fall back to plain DNS.
+%%
+query_nss(Q, [], Timer, Retry, I, S, Reason, RetryNSs) ->
+    %% End of name servers list, do a new retry
+    %% with the remaining name servers
+    query_retries(Q, lists:reverse(RetryNSs), Timer, Retry, I+1, S, Reason);
+query_nss(#q{edns = undefined}=Q, NSs, Timer, Retry, I, S, Reason, RetryNSs) ->
+    query_nss_dns(Q, NSs, Timer, Retry, I, S, Reason, RetryNSs);
+query_nss(Q, NSs, Timer, Retry, I, S, Reason, RetryNSs) ->
+    query_nss_edns(Q, NSs, Timer, Retry, I, S, Reason, RetryNSs).
 
 query_nss_edns(
-  #q{options=#options{udp_payload_size=PSz}=Options,edns={Id,Buffer}}=Q,
-  [{IP,Port}=NS|NSs]=NSs0, Timer, Retry, I, S0, ErrNSs) ->
-    {S,Res}=Reply =
-	query_ns(S0, Id, Buffer, IP, Port, Timer, Retry, I, Options, PSz),
-    case Res of
-	timeout -> {S,{error,timeout}}; % Bailout timeout
-	{ok,_} -> Reply;
-	{error,{nxdomain,_}} -> Reply;
-	{error,{E,_}} when E =:= qfmterror; E =:= notimp; E =:= servfail;
-			   E =:= badvers ->
-	    query_nss_dns(Q, NSs0, Timer, Retry, I, S, ErrNSs);
-	{error,E} when E =:= fmt; E =:= enetunreach; E =:= econnrefused ->
-	    query_nss(Q, NSs, Timer, Retry, I, S, [NS|ErrNSs]);
-	_Error ->
-	    query_nss(Q, NSs, Timer, Retry, I, S, ErrNSs)
+  #q{options =
+         #options{
+            udp_payload_size = PSz}=Options,
+     edns = {Id,Buffer}}=Q,
+  [NsSpec|NSs], Timer, Retry, I, S_0, Reason, RetryNSs) ->
+    %%
+    {IP,Port} = NS = servfail_retry_wait(NsSpec),
+    {S,Result} =
+	query_ns(
+          S_0, Id, Buffer, IP, Port, Timer, Retry, I, Options, PSz),
+    case Result of
+	{error,{E,_}}
+          when E =:= qfmterror;
+               E =:= notimp;
+               E =:= servfail;
+               E =:= badvers ->
+            %% The server did not like that.
+            %% Ignore that error and try plain DNS.
+            %%
+            %% We ignore the servfail_retry_timeout here,
+            %% assuming that if the servfail was due to us using EDNS,
+            %% a DNS query might work, therefore we do not
+            %% count this failure as a try.
+	    query_nss_dns(
+              Q, [NS|NSs], Timer, Retry, I, S, Reason, RetryNSs);
+        _ ->
+	    query_nss_result(
+              Q, NSs, Timer, Retry, I, S, Reason, RetryNSs, NS, Result)
     end.
 
 query_nss_dns(
-  #q{dns=Qdns}=Q0,
-  [{IP,Port}=NS|NSs], Timer, Retry, I, S0, ErrNSs) ->
-    #q{options=Options,dns={Id,Buffer}}=Q =
+  #q{dns = Qdns}=Q_0,
+  [NsSpec|NSs], Timer, Retry, I, S_0, Reason, RetryNSs) ->
+    %%
+    {IP,Port} = NS = servfail_retry_wait(NsSpec),
+    #q{options = Options,
+       dns = {Id,Buffer}}=Q =
 	if
-	    is_function(Qdns, 0) -> Q0#q{dns=Qdns()};
-	    true -> Q0
+	    is_function(Qdns, 0) -> Q_0#q{dns=Qdns()};
+	    true -> Q_0
 	end,
-    {S,Res}=Reply =
+    {S,Result} =
 	query_ns(
-	  S0, Id, Buffer, IP, Port, Timer, Retry, I, Options, ?PACKETSZ),
-    case Res of
-	timeout -> {S,{error,timeout}}; % Bailout timeout
-	{ok,_} -> Reply;
-	{error,{E,_}} when E =:= nxdomain; E =:= qfmterror -> Reply;
-	{error,E} when E =:= fmt; E =:= enetunreach; E =:= econnrefused ->
-	    query_nss(Q, NSs, Timer, Retry, I, S, [NS|ErrNSs]);
-	_Error ->
-	    query_nss(Q, NSs, Timer, Retry, I, S, ErrNSs)
+	  S_0, Id, Buffer, IP, Port, Timer, Retry, I, Options, ?PACKETSZ),
+    query_nss_result(
+      Q, NSs, Timer, Retry, I, S, Reason, RetryNSs, NS, Result).
+
+
+%% servfail_retry NsSpec handling.
+%%
+%% A NsSpec is either a NS = {IP, Port},
+%% or for a servfail_retry_timeout, the nameserver wrapped
+%% in a tuple with the earliest time to contact
+%% the nameserver again.
+%%
+%% When unwrapping; wait until it is time before returning
+%% the nameserver.
+
+%% Wrap with retry time
+servfail_retry_time(RetryTimeout, NS) ->
+    {servfail_retry, time(RetryTimeout), NS}.
+
+%% Unwrap and wait
+servfail_retry_wait(NsSpec) ->
+    case NsSpec of
+        {servfail_retry, Time, NS} ->
+            wait(timeout(Time)),
+            NS;
+        {_IP,_Port} = NS->
+            NS
     end.
 
+
+query_nss_result(Q, NSs, Timer, Retry, I, S, Reason, RetryNSs, NS, Result) ->
+    case Result of
+	{ok,_} ->
+            _ = udp_close(S),
+            Result;
+	timeout -> % Out of total time timeout
+            query_retries_error(S, Reason); % The best reason we have
+	{error,{nxdomain=E,_}} ->
+            query_retries_error(S, E); % Definite answer
+	{error,{E,_}=NewReason}
+          when E =:= qfmterror;
+               E =:= notimp;
+               E =:= refused;
+               E =:= badvers;
+               E =:= unknown ->
+            %% The server did not like that.
+            %% Do not retry this server since
+            %% it will not answer differently on the next retry.
+	    query_nss(Q, NSs, Timer, Retry, I, S, NewReason, RetryNSs);
+	{error,E=NewReason}
+          when E =:= formerr;
+               E =:= enetunreach;
+               E =:= econnrefused ->
+            %% Could not decode answer, or network problem.
+            %% Do not retry this server.
+	    query_nss(Q, NSs, Timer, Retry, I, S, NewReason, RetryNSs);
+	{error,timeout} -> % Query timeout
+            %% Try next server, may retry this server
+	    query_nss(Q, NSs, Timer, Retry, I, S, Reason, [NS|RetryNSs]);
+        {error,{servfail,_}=NewReason} ->
+            RetryTimeout = Q#q.options#options.servfail_retry_timeout,
+            case inet:timeout(RetryTimeout, Timer) of
+                RetryTimeout ->
+                    NsSpec = servfail_retry_time(RetryTimeout, NS),
+                    query_nss(
+                      Q, NSs, Timer, Retry, I, S, NewReason,
+                      [NsSpec|RetryNSs]);
+                _ ->
+                    %% No time for a new retry with this server
+                    %% - do not retry this server
+                    query_nss(
+                      Q, NSs, Timer, Retry, I, S, NewReason, RetryNSs)
+            end;
+	{error,NewReason} ->
+            %% NewReason =
+            %%     {error,badid} |
+            %%     {error,{noquery,Msg}} |
+            %%     {error,OtherSocketError}
+            %% Try next server, may retry this server
+	    query_nss(Q, NSs, Timer, Retry, I, S, NewReason, [NS|RetryNSs])
+    end.
+
+query_retries_error(S, Reason) ->
+    _ = udp_close(S),
+    {error, Reason}.
+
+
 query_ns(S0, Id, Buffer, IP, Port, Timer, Retry, I,
 	 #options{timeout=Tm,usevc=UseVC,verbose=Verbose},
 	 PSz) ->
     case UseVC orelse iolist_size(Buffer) > PSz of
 	true ->
 	    TcpTimeout = inet:timeout(Tm*5, Timer),
-	    {S0,query_tcp(TcpTimeout, Id, Buffer, IP, Port, Verbose)};
+	    {S0,
+             query_tcp(TcpTimeout, Id, Buffer, IP, Port, Verbose)};
 	false ->
 	    case udp_open(S0, IP) of
 		{ok,S} ->
-		    Timeout =
+		    UdpTimeout =
 			inet:timeout( (Tm * (1 bsl I)) div Retry, Timer),
-		    {S,
-		     case query_udp(
-			    S, Id, Buffer, IP, Port, Timeout, Verbose) of
+		     case
+                         query_udp(
+                           S, Id, Buffer, IP, Port, UdpTimeout, Verbose)
+                     of
 			 {ok,#dns_rec{header=H}} when H#dns_header.tc ->
 			     TcpTimeout = inet:timeout(Tm*5, Timer),
-			     query_tcp(
-			       TcpTimeout, Id, Buffer, IP, Port, Verbose);
-			 Reply -> Reply
-		     end};
+			     {S,
+                              query_tcp(
+                                TcpTimeout, Id, Buffer, IP, Port, Verbose)};
+			{error, econnrefused} = Err ->
+                            ok = udp_close(S),
+	                    {#sock{}, Err};
+			Reply -> {S, Reply}
+		     end;
 		Error ->
 		    {S0,Error}
 	    end
@@ -976,10 +1064,10 @@ decode_answer(Answer, Id, Verbose) ->
 		?NOTIMP   -> {error,{notimp,Msg}};
 		?REFUSED  -> {error,{refused,Msg}};
 		?BADVERS  -> {error,{badvers,Msg}};
-		_ -> {error,{unknown,Msg}}
+		_ ->         {error,{unknown,Msg}}
 	    end;
-	Error ->
-	    ?verbose(Verbose, "Got reply: ~p~n", [Error]),
+	{error, formerr} = Error ->
+	    ?verbose(Verbose, "Got reply: decode format error~n", []),
 	    Error
     end.
 
@@ -1057,10 +1145,32 @@ dns_msg(Msg) ->
 	    {Type,dns_msg(Fields)}
     end.
 
--compile({inline, [now_ms/2]}).
-now_ms(Int1, Int0) ->
-    Int1 - Int0.
 
--compile({inline, [time_now/0]}).
-time_now() ->
-	erlang:monotonic_time(1000).
+
+-compile({inline, [time/1, timeout/1, wait/1]}).
+
+%% What Time is the Timeout? [ms]
+%%
+time(Timeout) ->
+    erlang:monotonic_time(1000) + Timeout.
+
+%% How long Timeout to Time? [ms] >= 0
+%%
+timeout(Time) ->
+    TimeNow = erlang:monotonic_time(1000),
+    if
+        TimeNow < Time ->
+            Time - TimeNow;
+        true ->
+            0
+    end.
+
+%% receive after Timeout but do not yield for 0
+%%
+wait(0) ->
+    ok;
+wait(Timeout) ->
+    receive
+    after Timeout ->
+            ok
+    end.
